-------------------------------------------------------------------------------
--                       storage_pool_handler.ads                            --
--                                                                           --
-- Purposes: Manages memory resources. As a result memory leakage problem    --
--           will be thing of the past.                                      --
--                                                                           --
-- Anh Vo - 01 March 2000                                                    --
--                                                                           --
-- C/C++ is for money. Java is for fame and money. Ada is for reliability    --
--                                                                           -- 
-------------------------------------------------------------------------------

with Ada.Exceptions;
with Ada.Text_Io;
with System.Storage_Elements;
with System.Address_To_Access_Conversions;

package body Storage_Pool_Handler is

   use Ada;
   use Text_Io;
   use type System.Address;
   use type System.Storage_Elements.Storage_Count;

   Package_Name : constant String := "Storage_Pool_Handler.";


   -- used by General Pool
   type Address_Linker is 
      record
         Next_Address  : System.Address := System.Null_Address;
         Size_Elements : System.Storage_Elements.Storage_Count;
         Alignment     : System.Storage_Elements.Storage_Count;
      end record;
   
   -- storage elements needed by General Pool
   Link_Storage : constant System.Storage_Elements.Storage_Count :=
                             Address_Linker'Max_Size_In_Storage_Elements;

   package Address_To_Access is new 
                     System.Address_To_Access_Conversions (Address_Linker);

   
   -- used by detailed pool
   type Holder is
      record
         Next_Address : System.Address := System.Null_Address;
      end record;
   
   -- storage elements needed by Detailed Pool
   Holder_Storage : constant System.Storage_Elements.Storage_Count :=
                                      Holder'Max_Size_In_Storage_Elements;

   package Addr_To_Acc is new System.Address_To_Access_Conversions (Holder);
   

   function Aligned (
              Size      : System.Storage_Elements.Storage_Count;
              Alignment : System.Storage_Elements.Storage_Count)
                            return System.Storage_Elements.Storage_Count is
   begin
      if Size rem Alignment /= 0 then
         return Size + Alignment - (Size mod Alignment);
      else
         return Size;
      end if;
   end Aligned;


   procedure Allocate (
         Pool            : in out General_Pool;
         Storage_Address :    out System.Address;
         Size_In_Storage_Elements : in System.Storage_Elements.Storage_Count;
         Alignment       : in System.Storage_Elements.Storage_Count) is
              
      Previous : System.Address := System.Null_Address;
      Current  : System.Address := Pool.Addr_Head;
      Desired_Size : System.Storage_Elements.Storage_Count;

   begin   -- Allocate
     
      Pool.Counter := Pool.Counter + 1;

      Desired_Size := Aligned (Size_In_Storage_Elements, Alignment);
      if Desired_Size < Link_Storage then
         Desired_Size := Link_Storage;
      end if;
 
      while Current /= System.Null_Address loop
         if Address_To_Access.To_Pointer (Current).Size_Elements = 
                                                         Desired_Size and then
             Address_To_Access.To_Pointer (Current).Alignment = Alignment then

            -- matched block is found
            if Previous = System.Null_Address then
               Pool.Addr_Head := 
                 Address_To_Access.To_Pointer (Pool.Addr_Head).Next_Address;
            else
               -- adjust the link address to exclude the current one
               Address_To_Access.To_Pointer (Previous).Next_Address := 
                 Address_To_Access.To_Pointer (Current).Next_Address;
            end if;

            Storage_Address := Current;
            return;    -- mission complete
            
         else   -- check the next address of the link
         
            Previous := Current;
            Current := Address_To_Access.To_Pointer (Current).Next_Address;
         end if;            
      end loop;
      
      -- nothing found from storage reuse, wrap elements from pool     
      if Pool.Size - Pool.Addr_Index + 1 >= Desired_Size then
         Storage_Address := Pool.Data (Pool.Addr_Index)'Address;
         Pool.Addr_Index := Pool.Addr_Index + Desired_Size;
      else      
         Exceptions.Raise_Exception (Storage_Pool_Error'Identity,
            Package_Name & "Allocate: " & "Storage exhausted for user " & 
                                                          Pool.User.all);
      end if;
 
   end Allocate;

   procedure Deallocate (
         Pool            : in out General_Pool;
         Storage_Address : in     System.Address;
         Size_In_Storage_Elements : in System.Storage_Elements.Storage_Count;
         Alignment       : in System.Storage_Elements.Storage_Count) is
         
      New_Object : Address_To_Access.Object_Pointer;
      Desired_Size : System.Storage_Elements.Storage_Count;

   begin   -- Deallocate
   
      -- it never drops below 0
      Pool.Counter := Pool.Counter - 1;

      -- defragmenting the storage pool
      if Pool.Counter = 0 then
         Pool.Addr_Index := 1;
         Pool.Addr_Head := System.Null_Address;
         return;
      end if;

      New_Object := Address_To_Access.To_Pointer (Storage_Address);
      
      Desired_Size := Aligned (Size_In_Storage_Elements, Alignment);
      if Desired_Size < Link_Storage then
         Desired_Size := Link_Storage;
      end if;      

      New_Object.all := (Next_Address => Pool.Addr_Head,
                         Size_Elements => Desired_Size,
                         Alignment => Alignment);
        
      Pool.Addr_Head := Storage_Address;
        
   end Deallocate;


   function Storage_Size (Pool : General_Pool)
                            return System.Storage_Elements.Storage_Count is
   begin
      return Pool.Size;
   end Storage_Size;

  
   function Client_Id (Pool : General_Pool) return String is
   begin
      return Pool.User.all;
   end Client_Id;

   function Storage_Available (Pool : General_Pool)
                      return System.Storage_Elements.Storage_Count is
   begin
      return Pool.Size - Pool.Addr_Index + 1;
   end Storage_Available;

   function Largest_Block (Pool : General_Pool)
                            return System.Storage_Elements.Storage_Count is
      Free_Block : System.Storage_Elements.Storage_Count := 0;
      Addr_Index : System.Address := Pool.Addr_Head;
      The_Object : Address_To_Access.Object_Pointer;
   begin
      while Addr_Index /= System.Null_Address loop
         The_Object := Address_To_Access.To_Pointer (Addr_Index);
         
         if The_Object.Size_Elements > Free_Block then
            Free_Block := The_Object.Size_Elements;
         else
            Addr_Index := 
              Address_To_Access.To_Pointer (Addr_Index).Next_Address;
         end if;
      end loop;
      
      if Free_Block >= Pool.Size - Pool.Addr_Index + 1 then
         return Free_Block;
      else
         return Pool.Size - Pool.Addr_Index + 1; 
      end if;
   end Largest_Block;

   Procedure Initialize (Pool : in out General_Pool) is
   begin
      Put_Line (Pool.User.all & " user has the pool size of" &
        System.Storage_Elements.Storage_Count'Image (Pool.Size));
   end Initialize;

   Procedure Finalize (Pool : in out General_Pool) is
      Free_Count : System.Storage_Elements.Storage_Count := 0;
      Addr_Index : System.Address := Pool.Addr_Head;
      The_Object : Address_To_Access.Object_Pointer;
   begin
      while Addr_Index /= System.Null_Address loop
         The_Object := Address_To_Access.To_Pointer (Addr_Index);
         Free_Count := Free_Count + The_Object.Size_Elements;
         Addr_Index := Address_To_Access.To_Pointer (Addr_Index).Next_Address;
      end loop;
      
      if Free_Count /= Pool.Addr_Index - 1 then 
         Put_Line ("Memory Leakage Detected! " & Pool.User.all & 
                      "'s codes had memory leakage of " &
           System.Storage_Elements.Storage_Count'Image (
              Pool.Addr_Index - 1 - Free_count) & " storage elements.");
      else
         Put_Line ("All memory was returned to " & 
                                 Pool.User.all & "'s pool successfully.");
      end if;

   end Finalize;


   ----------------------------------------------------------------------------
   -- Detailed memory blocks size are handled by this pool                   --
   ----------------------------------------------------------------------------
   procedure Allocate (
         Pool            : in out Detailed_Pool;
         Storage_Address :    out System.Address;
         Size_In_Storage_Elements : in System.Storage_Elements.Storage_Count;
         Alignment       : in System.Storage_Elements.Storage_Count) is
   begin   

      if Pool.Head_Keeper /= System.Null_Address then
         Storage_Address := Pool.Head_Keeper;
         Pool.Head_Keeper :=
                    Addr_To_Acc.To_Pointer (Pool.Head_Keeper).Next_Address;
      else
         Exceptions.Raise_Exception (Storage_Pool_Error'Identity,
            "Storage exhausted in " & Package_Name & "Allocate" & 
            " used by " & Pool.User.all);
      end if;
   end Allocate;

   procedure Deallocate (
         Pool                     : in out Detailed_Pool;
         Storage_Address          : in System.Address;
         Size_In_Storage_Elements : in System.Storage_Elements.Storage_Count;
         Alignment : in System.Storage_Elements.Storage_Count ) is
   begin
      Addr_To_Acc.To_Pointer (Storage_Address).Next_Address := Pool.Head_Keeper;
      Pool.Head_Keeper := Storage_Address;

   end Deallocate;

   function Storage_Size (Pool : in Detailed_Pool ) 
                            return System.Storage_Elements.Storage_Count is
   begin
      return Pool.Size;
   end Storage_Size;
   
   function User_Id (Pool : Detailed_Pool) return String is
   begin
      return Pool.User.all;
   end User_Id;

   function Block_Available (Pool : Detailed_Pool) return Natural is
      Count : Natural := 0;
      Temp_Addr : System.Address := Pool.Head_Keeper;
   begin
      while Temp_Addr /= System.Null_Address loop
         Count := Count + 1;
         Temp_Addr := Addr_To_Acc.To_Pointer (Temp_Addr).Next_Address;
      end loop;
      return Count;
   end Block_Available; 

   procedure Initialize (Pool : in out Detailed_Pool) is
      Temp_Addr : System.Address;
      Addr_Index : System.Storage_Elements.Storage_Count := 1;
      
   begin   -- Initialize
   
      if Pool.Max < Holder_Storage then
         Exceptions.Raise_Exception (Storage_Pool_Error'Identity, 
           "Not appropriate for Detailed Pool. Use General Pool instead");
      end if;
      
      while Pool.Size >= Addr_Index + Pool.Max - 1 loop
         Temp_Addr := Pool.Data (Addr_Index)'Address;
         Addr_Index := Addr_Index + Pool.Max;
         Addr_To_Acc.To_Pointer (Temp_Addr).Next_Address := Pool.Head_keeper;
         Pool.Head_Keeper := Temp_Addr;
         Pool.Count := Pool.Count + 1;
      end loop;
           
      Put_Line (Pool.User.all & " user has" & 
        Integer'Image (Pool.Count) & " blocks with the Max block of" & 
        System.Storage_Elements.Storage_Count'Image (Pool.Max) & 
                 " elements for the pool size of" &
        System.Storage_Elements.Storage_Count'Image (Pool.Size));
   end Initialize;

   procedure Finalize (Pool : in out Detailed_Pool) is
   begin
      if Block_Available (Pool) /= Pool.Count then
         Put_Line ("Memory Leakage Detected! " & Pool.User.all & 
                             "'s codes had memory leakage of" &
           Natural'Image ((Pool.Count - Block_Available (Pool)) * 
                             Natural (Pool.Max)) & " storage elements.");
      else
         Put_Line ("All memory was returned to " & 
                                 Pool.User.all & "'s pool successfully.");
      end if;
   end Finalize;
   
end Storage_Pool_Handler;